// Copyright (c) 2000-2001 Brad Hughes <bhughes@trolltech.com>
//
// Use, modification and distribution is allowed without limitation,
// warranty, or liability of any kind.
//
#include <QtGui>
#include <QObject>
#include <QStringList>
#include <QApplication>


#include "constants.h"
#include "buffer.h"
#include "output.h"
#include "visual.h"
#include "factory.h"
#include "equ/iir.h"
#include "decoder.h"



Decoder::Decoder(QObject *parent, DecoderFactory *d, QIODevice *i, Output *o)
        : QThread(parent), fctry(d), in(i), out(o), blksize(0),m_eqInited(FALSE),
          m_useEQ(FALSE)
{
    int b[] = {0,0,0,0,0,0,0,0,0,0}; 
    setEQ(b, 0);
}

Decoder::~Decoder()
{
    fctry = 0;
    in = 0;
    out = 0;
    blksize = 0;
}


void Decoder::setInput(QIODevice *i)
{
    mutex()->lock()
    ;
    in = i;
    mutex()->unlock();
}


void Decoder::setOutput(Output *o)
{
    mutex()->lock()
    ;
    out = o;
    mutex()->unlock();
}



/*void Decoder::dispatch(const DecoderEvent &e)
{
    QObject *object = listeners.first();
    while (object) {
	//QApplication::postEvent(object, new DecoderEvent(e));
	object = listeners.next();
    }
}*/


void Decoder::error(const QString &e)
{
    QObject *object = listeners.first();
    /*while (object) {
    QString *str = new QString(e.utf8());
    //QApplication::postEvent(object, new DecoderEvent(str));
    object = listeners.next();
    }*/
}


void Decoder::addListener(QObject *object)
{
    if  (listeners.indexOf(object) == -1)
        ;
    listeners << object;
}


void Decoder::removeListener(QObject *object)
{
    listeners.removeAll(object);
}


// static methods

static QList<DecoderFactory*> *factories = 0;
static QStringList files;

static void checkFactories()
{
    if (! factories)
    {
        files.clear();
        factories = new QList<DecoderFactory *>;

        QDir pluginsDir (qApp->applicationDirPath());
        pluginsDir.cdUp();
        pluginsDir.cd("Plugins/Input");
        foreach (QString fileName, pluginsDir.entryList(QDir::Files))
        {
            QPluginLoader loader(pluginsDir.absoluteFilePath(fileName));
            QObject *plugin = loader.instance();
            if (loader.isLoaded())
            {
                qDebug("Decoder: plugin loaded - %s", qPrintable(fileName));
            }
            DecoderFactory *factory = 0;
            if (plugin)
                factory = qobject_cast<DecoderFactory *>(plugin);

            if (factory)
            {
                Decoder::registerFactory(factory);
                files << pluginsDir.absoluteFilePath(fileName);
            }
        }
    }
}


QStringList Decoder::all()
{
    checkFactories();

    QStringList l;
    DecoderFactory *fact;
    foreach(fact, *factories)
    {
        l << fact->description();
    }
    return l;
}

QStringList Decoder::decoderFiles()
{
    checkFactories();
    return files;
}


bool Decoder::supports(const QString &source)
{
    checkFactories();

    DecoderFactory *fact = factories->first();
    foreach(fact, *factories)
    {
        if (fact->supports(source))
        {
            return TRUE;
        }
    }

    return FALSE;
}


void Decoder::registerFactory(DecoderFactory *fact)
{
    factories->append(fact);
}


Decoder *Decoder::create(QObject *parent, const QString &source,
                         QIODevice *input,
                         Output *output,
                         bool deletable)
{
    Decoder *decoder = 0;

    DecoderFactory *fact = Decoder::findFactory(source);
    if(fact)
    {
        decoder = fact->create(parent, input, output, deletable);
    }
    qDebug ("check ends");
    return decoder;
}

DecoderFactory *Decoder::findFactory(const QString& source)
{
    checkFactories();
    DecoderFactory *fact;
    foreach(fact, *factories)
    {
        if (fact->supports(source))
        {
            return fact;
        }
    }
    qDebug("Decoder: unable to find factory");
    return 0;
}

FileTag *Decoder::createTag(const QString& source)
{
    DecoderFactory *fact = Decoder::findFactory(source);
    if(fact)
    {
        return fact->createTag(source);
    }
    return 0;
}

QString Decoder::filter()
{
    QString allflt("All Supported Bitstreams (");
    QString flt;

    checkFactories();
    DecoderFactory *fact;
    for (int i = 0; i<factories->size(); ++i)
    {
        fact = (*factories)[i];
        allflt +=fact->filter().toLower() +" ";
        flt += fact->description() + " (" + fact->filter().toLower() + ")";
        if (i+1<factories->size())
        {
            fact = (*factories)[i+1];
            flt += ";;";
        }
    }

    allflt += ");;";

    return allflt + flt;
}

QStringList Decoder::nameFilters()
{
    checkFactories();
    QStringList filters;
    DecoderFactory *fact;
    foreach(fact, *factories)
    {
        filters << fact->filter().split(" ", QString::SkipEmptyParts);
    }
    return filters;
}

QList<DecoderFactory*> *Decoder::decoderFactories()
{
    checkFactories();
    return factories;
}

void Decoder::dispatch(const DecoderState &st)
{
    m_state = st;
    emit stateChanged(&m_state);
}


ulong Decoder::produceSound(char *data, ulong output_bytes, ulong bitrate, int nch)
{
    ulong sz = output_bytes < blksize ? output_bytes : blksize;
    Buffer *b = output()->recycler()->get();

    if(!m_eqInited)
    {
        init_iir();
        m_eqInited = TRUE;
    }
    if(m_useEQ)
       iir(data,sz,nch);

    memcpy(b->data, data, sz);
    if (sz != blksize)
        memset(b->data + sz, 0, blksize - sz);

    b->nbytes = blksize;
    b->rate = bitrate;
    output()->recycler()->add();

    output_bytes -= sz;
    memmove(data, data + sz, output_bytes);
    return sz;
}

void Decoder::setEQ(int bands[10], int preamp)
{
   set_preamp(0, 1.0 + 0.0932470 *preamp + 0.00279033 * preamp * preamp);
   set_preamp(1, 1.0 + 0.0932470 *preamp + 0.00279033 * preamp * preamp);
   for (int i=0; i<=10; ++i)
   {
      int value = bands[i];
      set_gain(i,0, 0.03*value+0.000999999*value*value);
      set_gain(i,1, 0.03*value+0.000999999*value*value);
   }
}
